@@ -3,7 +3,7 @@ from sklearn.manifold import TSNE
 from umap import UMAP  # takes a while to load
 
 ALGORITHMS = {
-    "TruncatedSVD": TruncatedSVD,
-    "UMAP": UMAP,
-    "TSNE": TSNE,
+    TruncatedSVD.__name__: TruncatedSVD,
+    UMAP.__name__: UMAP,
+    TSNE.__name__: TSNE,
 }
