@@ -1,11 +1,12 @@
 from cProfile import label
 import logging
-import os
+import warnings
 import os.path
 
 import anndata
 import scanpy as sc
 
+import tqdm
 import pandas as pd
 import numpy as np
 
@@ -20,11 +21,9 @@ from sleep_models.constants import (
     TRAIN_SIZE,
 )
 
-from sleep_models.utils.logging import setup_logging
 import sleep_models.utils.data as data_utils
 from sleep_models.utils.data import split_dataset
 
-logger = logging.getLogger("sleep_models.preprocessing")
 logger = logging.getLogger(__name__)
 
 # Creating Test DataSets using sklearn.datasets.make_blobs
@@ -33,16 +32,19 @@ logger = logging.getLogger(__name__)
 def read_h5ad(
     file,
     *args,
-    highly_variable_genes=False,
     raw=False,
+    highly_variable_genes=False,
     exclude_genes_file=None,
     **kwargs,
 ):
 
     adata = _read_h5ad(file, *args, **kwargs)
+    if raw:
+        adata = _use_raw_anndata(adata)
+
     if highly_variable_genes:
         if not "highly_variable" in adata.var.columns:
-            highly_variable_genes(
+            _highly_variable_genes(
                 adata,
                 min_mean=MIN_MEAN_HVG,
                 max_mean=MAX_MEAN_HVG,
@@ -51,9 +53,6 @@ def read_h5ad(
 
         adata = adata[:, adata.var.index[adata.var["highly_variable"]]]
 
-    if raw:
-        adata = _use_raw_anndata(adata)
-
     if exclude_genes_file is not None:
         logger.info(f"Removing genes listed in {exclude_genes_file}")
         adata = remove_genes(adata, exclude_genes_file)
@@ -63,10 +62,10 @@ def read_h5ad(
 
 def load_adata(
     h5ad_input,
-    cluster,
+    cluster=None,
     exclude_genes_file=None,
-    highly_variable_genes=True,
-    logger=None,
+    raw=False,
+    highly_variable_genes=False
 ):
     """
     Loads an anndata.AnnData cached to an h5ad file
@@ -86,16 +85,22 @@ def load_adata(
         adata (anndata.Anndata)
     """
 
-    if logger is None:
-        logger = logging.getLogger(__name__)
-
     logger.info(f"Reading {h5ad_input} to memory")
-    print(f"Reading {h5ad_input} to memory")
     adata = read_h5ad(
         h5ad_input,
+        raw=raw,
         highly_variable_genes=highly_variable_genes,
         exclude_genes_file=exclude_genes_file,
     )
+
+
+    if "CellType" not in adata.obs.columns:
+        # NOTE
+        # The input .h5ad provided by joana contains this column
+        # which should be called CellType throughout the pipeline from now on
+        # If you adata already contains CellType, this code is ignored
+        adata.obs["CellType"]=adata.obs["Cluster_ID_res8_080822"]
+
     if cluster is not None:
         adata = adata[adata.obs["CellType"] == cluster]
     return adata
@@ -228,6 +233,7 @@ class Pipeline:
 
         # simplify y if a mapping is pased:
         # useful to group values of the target into broader categories
+
         if label_mapping is not None:
             y = self.simplify_y(y, mapping=label_mapping)
 
@@ -252,7 +258,10 @@ class Pipeline:
             keep_all_cells = np.arange(X.shape[0])
 
         # encode y as fit for ML
-        y = pd.DataFrame(self.encode_y(y))
+        if self._encoding is not None:
+            y = pd.DataFrame(self.encode_y(y))
+        else:
+            y = pd.DataFrame(y)
         X.columns = adata.var.index
         y.index = X.index
 
@@ -263,25 +272,25 @@ class Pipeline:
         return X, y
 
     @staticmethod
-    def remove_mean(X_train, y_train, X_test, y_test):
+    def remove_mean(X_train, y_train, X_val, y_val):
         feat_mean = X_train.mean(axis=1)
         X_train -= feat_mean
-        X_test -= feat_mean
-        return X_train, y_train, X_test, y_test
+        X_val -= feat_mean
+        return X_train, y_train, X_val, y_val
 
 
-def scale_x(X_train, X_test):
+def scale_x(X_train, X_val):
     """
-    
+
     """
     mean = X_train.mean(axis=0)
     X_train -= mean
     std = X_train.std(axis=0)
     X_train /= std
 
-    X_test = X_test - mean / std
+    X_val = X_val - mean / std
 
-    X_train, X_test
+    X_train, X_val
 
 
 def remove_genes(adata, exclude_genes_file):
@@ -306,7 +315,6 @@ def keep_cells_from_this_background(adata, background):
     Arguments:
         adata (anndata.AnnData): Input anndata to be filtered
         background (pd.DataFrame):
-
     """
 
     assert "louvain_res" in background.columns
@@ -318,34 +326,34 @@ def keep_cells_from_this_background(adata, background):
     return adata
 
 
-def shuffle_adata(adata, filename, pinned_columns=[]):
+def shuffle_adata(adata, filename=None, pinned_columns=[]):
     """
     The result of the shuffling is a new adata where the rows of the counts matrix (adata.X)
     are shuffled, so the reads for cell 1 are assigned to any other random cell
     This is useful when simulating null hypotheses
     If pinned_columns is passed, this column in the obs table are NOT shuffled, i.e. cells that had a given value
     in the column will be shuffled only with cells with the same value
+
+
+    Arguments:
+
+    * adata (anndata.AnnData)
+    * filename (str): If not None, the shuffled adata is saved to this path
+    * pinned_columns (list): Columns in the obs table that are not shuffled i.e.
+    all counts belonging to cells with value A on one of the pinned columns will still
+    belong to the same group of cells
+
+    Return:
+
+    * adata_shuffled (anndata.Anndata) 
     """
 
     adatas = []
     if pinned_columns:
-        pins = []
-        for i in range(adata.shape[0]):
-            pin = []
-            for column in pinned_columns:
-                pin.append(adata.obs.iloc[i][column])
-
-            pins.append(tuple(pin))
-
-        pins = set(pins)
-
-        for pin in pins:
-            selector = []
-            for i, k in enumerate(pin):
-                selector.append((adata.obs[pinned_columns[i]].values == k))
-
-            selector = np.stack(selector, axis=1).all(axis=1)
 
+        selectors = None
+        selectors = [element[1].index for element in adata.obs[pinned_columns].groupby(pinned_columns)]
+        for selector in selectors:
             adata_shuffled = adata[selector, :].copy()
             adata_shuffled.obs = adata_shuffled.obs.sample(frac=1)
             adatas.append(adata_shuffled)
@@ -356,12 +364,9 @@ def shuffle_adata(adata, filename, pinned_columns=[]):
         adata_shuffled = adata.copy()
         adata_shuffled.obs = adata_shuffled.obs.sample(frac=1)
 
-    # adata_shuffled.obs.reset_index(inplace=True)
-    # adata_shuffled.obs["original_index"]=adata_shuffled.obs["index"]
-    # adata_shuffled.obs.drop("index", axis=1, inplace=True)
-
-    logger.info(f"Saving shuffled h5ad to disk at {filename}")
-    adata_shuffled.write_h5ad(filename)
+    if filename is not None:
+        logger.info(f"Saving shuffled h5ad to disk at {filename}")
+        adata_shuffled.write_h5ad(filename)
     return adata_shuffled
 
 
@@ -374,6 +379,7 @@ def get_bad_genes(batch_genes_file, exclude_genes_file):
     Combine both and return them
     """
     if batch_genes_file is not None:
+        warnings.warn("batch_genes_file is deprecated. Please add batch effect genes in the exclude_genes_file")
         batch_genes = pd.read_excel(batch_genes_file, sheet_name="all")[
             "gene"
         ].values.tolist()
@@ -465,7 +471,7 @@ def simulate_data(
 
 
     Returns: (np.ndarray, np.ndarray, np.ndarray, np.ndarray), i.e.
-    a tuple with the X and y of the training and test sets respectively
+    a tuple with the X and y of the training and validation sets respectively
     """
 
     label_mapping = data_utils.load_label_mapping(label_mapping)
@@ -498,18 +504,18 @@ def simulate_data(
     X = np.float32(X)
 
     if stratify:
-        X_train, y_train, X_test, y_test = split_dataset(
+        X_train, y_train, X_val, y_val = split_dataset(
             X, y, stratify=y, random_state=random_state, train_size=0.75
         )
     else:
-        X_train, y_train, X_test, y_test = split_dataset(
+        X_train, y_train, X_val, y_val = split_dataset(
             X, y, random_state=random_state, train_size=0.75
         )
 
     if mean_scale:
-        X_train, X_test = scale_x(X_train, X_test)
+        X_train, X_val = scale_x(X_train, X_val)
 
-    return {"datasets": (X_train, y_train, X_test, y_test), "encoding": None}
+    return {"datasets": (X_train, y_train, X_val, y_val), "encoding": None}
 
 
 def sort_encoding(label_mapping, encoding):
@@ -555,10 +561,12 @@ def load_data(
     highly_variable_genes=HIGHLY_VARIABLE_GENES,
     verbose=logging.WARNING,
     logfile=None,
+    template_file=None,
     label_mapping=None,
     trim=False,
     fraction=1.0,
     stratify=True,
+    shuffle=None,
 ):
 
     """
@@ -572,7 +580,7 @@ def load_data(
 
     Arguments:
 
-        h5ad_input (str): Path to an h5ad file containing an anndata.AnnData
+        h5ad_input (str, anndata.AnnData): Path to an h5ad file containing an anndata.AnnData, or an already loaded anndata.AnnData
             This AnnData must fulfil the following requirements
                 * .obs["CellType"] is defined
                 * .obs["Condition"] or .obs["Treatment"] are defined (depending on the selected target)
@@ -585,33 +593,43 @@ def load_data(
         highly_variable_genes (bool): If True, only the genes labeled as highly variable are used, otherwise all are used
             A gene is highly variable if it is found so in `sleep_models.preprocessing.read_h5ad`
 
-        
+
         trim (boolean): If True, all labels are downsampled to the least frequent label
         fraction (float): Fraction of the dataset to be used
 
     Returns: (np.ndarray, np.ndarray, np.ndarray, np.ndarray), i.e.
-    a tuple with the X and y of the training and test sets respectively
+    a tuple with the X and y of the training and validation sets respectively
     """
-
+        
     label_mapping = data_utils.load_label_mapping(label_mapping)
 
     if logfile is None:
         logfile = os.path.join(output, "logs", f"train_model_{cluster}.log")
 
-    logger, _ = setup_logging(verbose, logfile)
     logger.info(f"Training on cell type {cluster} starting!")
 
     data_preprocessor = Pipeline(
         model_properties=model_properties, target=target, random_state=random_state
     )
 
-    adata = load_adata(
-        h5ad_input,
-        cluster=cluster,
-        exclude_genes_file=exclude_genes_file,
-        highly_variable_genes=highly_variable_genes,
-        logger=logger,
-    )
+    if isinstance(h5ad_input, str):
+        adata = load_adata(
+            h5ad_input,
+            cluster=cluster,
+            exclude_genes_file=exclude_genes_file,
+        )
+    elif isinstance(h5ad_input, anndata.AnnData):
+        adata = h5ad_input
+
+
+    if shuffle is not None:
+        print("Shuffling adata")
+        adata = shuffle_adata(adata, filename=None)
+
+    if template_file:
+        template = data_utils.load_template(template_file)
+        adata = template_matching(adata, template)
+
 
     if fraction != 1.0:
         keep_n_cells = int(adata.shape[0] * fraction)
@@ -621,32 +639,46 @@ def load_data(
         adata, target=target, label_mapping=label_mapping, trim=trim
     )
 
-    if stratify:
-        X_train, y_train, X_test, y_test = split_dataset(
-            X, y, stratify=y, random_state=random_state, train_size=0.75
-        )
-    else:
-        X_train, y_train, X_test, y_test = split_dataset(
-            X, y, random_state=random_state, train_size=0.75
-        )
+    labels, counts = np.unique(np.argmax(y.values, 1), return_counts=True)
+    
+    if len(labels == 2) and np.min(counts) == 1:
+        logger.warning(f"One of the classes has only one instance. Skipping {cluster}")
+        return None
 
-    encoding = list(enumerate(data_preprocessor._encoder.categories_[0]))
-    encoding = {v[0]: v[1] for v in encoding}
+    try:
+        if stratify:
+            X_train, y_train, X_val, y_val = split_dataset(
+                X, y, stratify=y, random_state=random_state, train_size=0.75
+            )
+        else:
+            X_train, y_train, X_val, y_val = split_dataset(
+                X, y, random_state=random_state, train_size=0.75
+            )
+    
+    except Exception as error:
+        print(error)
+        import ipdb; ipdb.set_trace()
 
-    encoding = sort_encoding(label_mapping, encoding)
+    try:
+        encoding = list(enumerate(data_preprocessor._encoder.categories_[0]))
+        encoding = {v[0]: v[1] for v in encoding}
+        encoding = sort_encoding(label_mapping, encoding)
+
+    except AttributeError:
+        encoding=None
 
     data_utils.backup_dataset(
         cluster,
         X_train,
         y_train,
-        X_test,
-        y_test,
+        X_val,
+        y_val,
         data_preprocessor.original_data,
         encoding=encoding,
         output=output,
     )
     return {
-        "datasets": (X_train.values, y_train.values, X_test.values, y_test.values),
+        "datasets": (X_train.values, y_train.values, X_val.values, y_val.values),
         "encoding": encoding,
     }
 
@@ -719,6 +751,7 @@ def remove_genes_from_list(adata, discarded_genes):
     return adata
 
 
+
 def _read_h5ad(file, *args, **kwargs):
     """
     Read a cached anndata.AnnData using scanpy
@@ -727,7 +760,14 @@ def _read_h5ad(file, *args, **kwargs):
 
 
 def _highly_variable_genes(*args, **kwargs):
-    return sc.pp.highly_variable_genes(*args, **kwargs)
+    if "inplace" in kwargs:
+        inplace = kwargs.pop("inplace")
+        if not inplace:
+            logger.warning("highly variable genes will be computed inplace")
+
+    logger.debug("Computing highly variable genes")
+
+    return sc.pp.highly_variable_genes(*args, **kwargs, inplace=True)
 
 
 def _use_raw_anndata(adata):
@@ -743,7 +783,6 @@ def _use_raw_anndata(adata):
 def _concatenate_adatas(adatas):
     return anndata.concat(adatas)
 
-
 def _get_cluster(adata, louvain_resolution, idx):
     """
     set a single cell dataset
@@ -766,8 +805,8 @@ def _get_cluster(adata, louvain_resolution, idx):
 
     idx = str(idx)
     barcodes = adata.obs.loc[adata.obs[louvain_resolution] == idx].index.tolist()
-    logger.info(f"n cells = {len(barcodes)}")
-    logger.info(f"n genes = {adata.var.shape[0]}")
+    logger.info(f"{idx} - n cells = {len(barcodes)}")
+    logger.info(f"{idx} - n genes = {adata.var.shape[0]}")
     return barcodes
 
 
@@ -798,42 +837,4 @@ def _find_barcodes_belonging_to_clusters(adata, identifiers):
         logger.warning(f"{n_repeated_cells} cells appear in more than one cluster")
         matched_barcodes = list(set(matched_barcodes))
 
-    return matched_barcodes
-
-
-def assign_cell_type(adata, background):
-    """
-    Annotate on the CellType column of adata.obs
-
-    Arguments:
-
-        adata (anndata.AnnData): An Anndata object annotated with louvain_res and idx for each cell
-        background (pd.DataFrame): A data frame with columns [cluster,louvain_res,idx]
-
-    Returns:
-        adata with new column called CellType containing the cell type (cluster) annotation of each cell
-    """
-
-    clusters = [
-        (row["louvain_res"], str(row["idx"])) for i, row in background.iterrows()
-    ]
-
-    barcodes = [
-        _find_barcodes_belonging_to_clusters(adata, [clusters[i]])
-        for i in range(len(clusters))
-    ]
-
-    cluster_barcodes = list(zip(background["cluster"].values, barcodes))
-
-    cell_type = [None,] * adata.obs.shape[0]
-
-    for i in range(len(cluster_barcodes)):
-
-        cell_type_name, cell_type_barcodes = cluster_barcodes[i]
-        cell_type_pos = [barcode in cell_type_barcodes for barcode in adata.obs.index]
-        for j, status in enumerate(cell_type_pos):
-            if status:
-                cell_type[j] = cell_type_name
-
-    adata.obs["CellType"] = cell_type
-    return adata
+    return matched_barcodes
\ No newline at end of file
